package pt.uc.dei.wsst.gbt;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.jdom.Text;
import static pt.uc.dei.wsst.gbt.Util.isTraining;
import static pt.uc.dei.wsst.gbt.Util.logar;
import static pt.uc.dei.wsst.gbt.Util.vulne;
import static pt.uc.dei.wsst.gbt.Util.trace;

/**
 *
 * @author Nuno Antunes
 * @author Nuno Laranjeiro
 */
@Aspect
public class InjectionDetector {

    private static Map<String, Set<String>> sChecksumMap;
    private static Map<String, Set<String>> xpathChecksums;

    /**
     * @see java.sql.Statement#executeQuery(java.lang.String)
     * @see java.sql.Statement#executeUpdate
     * @see java.sql.Statement#execute
     * @see org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate#query
     * @see org.springframework.jdbc.core.JdbcTemplate#query
     * @see org.springframework.jdbc.core.simple.SimpleJdbcTemplate#query
     * @see org.springframework.jdbc.core.JdbcTemplate#queryForObject
     * @see org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate#update
     *
     *
     * @param proceedingJoinPoint
     * @return
     * @throws Throwable
     */
    @Around("call(* java.sql.Statement.executeQuery(..)) || call(* java.sql.Statement.executeUpdate(..)) || call(* java.sql.Statement.execute(..)) || call(* org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate.query(..)) || call(* org.springframework.jdbc.core.JdbcTemplate.query(..)) || call(* org.springframework.jdbc.core.simple.SimpleJdbcTemplate.query(..)) || call(* org.springframework.jdbc.core.JdbcTemplate.queryForObject(..)) || call(* org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate.update(..))")
    public Object aroundSQLInvocation(ProceedingJoinPoint proceedingJoinPoint)
            throws Throwable {
        logar("aroundSQLInvocation: " + proceedingJoinPoint);
        initMaps();
        boolean training = isTraining();
        String sql = (String) proceedingJoinPoint.getArgs()[0];
        trace("SQL: " + sql);
        String checksum = calculateSQLChecksum(sql);
        trace("SQL Checksum: " + checksum);
        if (training) {
            addChecksum(proceedingJoinPoint.getSourceLocation().toString(),
                    checksum, sChecksumMap);
        } else {
            Set<String> checksumList = sChecksumMap.get(proceedingJoinPoint.getSourceLocation().toString());
            if (checksumList == null) {
                vulne("Training period did not exercise this code point. sql: " + sql + " checksum: " + checksum + " location: " + proceedingJoinPoint.getSourceLocation());
            } else if (!checksumList.contains(checksum)) {
                vulne("SQL Injection attack detected. sql: " + sql + " checksum: " + checksum + " location: " + proceedingJoinPoint.getSourceLocation());
            }
        }
        return proceedingJoinPoint.proceed(proceedingJoinPoint.getArgs());
    }

    /**
     *
     * @see org.jdom.xpath.XPath#newInstance(java.lang.String)
     * @see org.jdom.xpath.XPath#selectNodes(java.lang.Object, java.lang.String)
     * @see org.jdom.xpath.XPath#selectSingleNode(java.lang.Object,
     *      java.lang.String)
     *
     *
     *
     * @param proceedingJoinPoint
     * @return
     * @throws java.lang.Throwable
     */
    @Around("call(* org.jdom.xpath.XPath.newInstance(..)) || call(* org.jdom.xpath.XPath.selectNodes(..)) || call(* org.jdom.xpath.XPath.selectSingleNode(..))")
    public Object aroundXPathInvocation(ProceedingJoinPoint proceedingJoinPoint)
            throws Throwable {
        logar("aroundXPathInvocation: " + proceedingJoinPoint);
        initMaps();
        boolean training = isTraining();
        Object[] args = proceedingJoinPoint.getArgs();
        String path = (String) args[args.length - 1];
        String checksum = calculateXPathChecksum(path);
        if (training) {
            addChecksum(proceedingJoinPoint.getSourceLocation().toString(),
                    checksum, xpathChecksums);
        } else {
            Set<String> checksumList = xpathChecksums.get(proceedingJoinPoint.getSourceLocation().toString());
            if (checksumList == null) {
                vulne("[XPath]Training period did not exercise this code point. path: " + path + " checksum: " + checksum + " location: " + proceedingJoinPoint.getSourceLocation());
            } else if (!checksumList.contains(checksum)) {
                vulne("XPath Injection attack detected. path: " + path + " checksum: " + checksum + " location: " + proceedingJoinPoint.getSourceLocation());
            }
        }
        return proceedingJoinPoint.proceed(proceedingJoinPoint.getArgs());
    }

    private static void initMaps() {
        if (sChecksumMap == null) {
            sChecksumMap = new ConcurrentHashMap<String, Set<String>>();
        }
        if (xpathChecksums == null) {
            xpathChecksums = new ConcurrentHashMap<String, Set<String>>();
        }
        trace("sChecksumMap: " + sChecksumMap.size());
        trace("xpathChecksums: " + xpathChecksums.size());
    }
    public static final String NOT_REPLACE = "MUST_NOT_REPLACE_THIS_N_U_L_L_FOUND";
    static Pattern NULL_FINDER = Pattern.compile("([a-zA-Z0-9](null))|((null)[a-zA-Z0-9])");

    private static String calculateSQLChecksum(String sql) {
        String result = sql.replaceAll("'[^']*'", "FOUND_A_STRING");
        result = result.replaceAll("\\((\\d+\\,)*\\d+\\)", "FOUND_A_LIST");
        result = result.replaceAll("-?\\d+", "FOUND_A_NUMBER");// -?
        result = result.replaceAll(" null ", " FOUND_A_NUMBER ");// -?
        int lastIndex = 0;
        StringBuilder newStr = new StringBuilder();
        for (Matcher matcher = NULL_FINDER.matcher(result); matcher.find();) {
            int group = matcher.group(1) != null ? 1 : 3;
            newStr.append(result.substring(lastIndex, matcher.start(group))).append(NOT_REPLACE);
            lastIndex = matcher.end(group) - 1;
        }
        if (lastIndex < result.length()) {
            newStr.append(result.substring(lastIndex));
        }
        result = newStr.toString().replaceAll("null", "FOUND_A_NUMBER").replaceAll(NOT_REPLACE, "null");
        return Text.normalizeString(result);
    }

    /**
     * Method copied from {@link #calculateSQLChecksum(java.lang.String) }.
     * Testing needed...
     *
     * @param path
     *            the path to generate checksum
     * @return the query checksum
     */
    private static String calculateXPathChecksum(String path) {
        String result = path.replaceAll("'[^']*'", "FOUND_A_STRING");
        result = result.replaceAll("\\((\\d+\\,)*\\d+\\)", "FOUND_A_LIST");
        result = result.replaceAll("-?\\d+", "FOUND_A_NUMBER");
        return Text.normalizeString(result);
    }

    private static void addChecksum(String sourceLocation, String checksum,
            Map<String, Set<String>> map) {
        Set<String> checksumList = map.get(sourceLocation);
        if (checksumList == null) {
            checksumList = new HashSet<String>();
            map.put(sourceLocation, checksumList);
        }
        if (checksumList.add(checksum)) {
            vulne("Found @ " + sourceLocation + " \t CheckSum: " + checksum);
        }
    }
}
